import subprocess

from tinydb import TinyDB, Query
from typing import Literal, Tuple

from .config import DB_FILENAME


class Check:
    """A command which is run on a given host with a certain expected output."""

    def __init__(self, host, username, command, expected, description=""):
        self.host = host
        self.username = username
        self.command = command
        self.expected = expected
        self.description = description

    def dump(self, short: bool = False) -> dict:
        """Returns a dictionary containing all relevant data."""
        data = {
            "host": self.host,
            "command": self.command,
            "expected": self.expected,
            "description": self.description,
        }
        if not short:
            status, stdout, stderr = self._get_status()
            data["status"] = status
            data["stdout"] = stdout
            data["stderr"] = stderr

        return data

    def _execute_ssh_command(self) -> str:
        """Executes an SSH command and returns stdout and stderr."""
        result = subprocess.run(
            [
                "ssh",
                "-l",
                self.username,
                self.host,
                "--",
                self.command,
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        def _decode(b):
            return b.decode("utf-8").strip() if b else ""

        return _decode(result.stdout), _decode(result.stderr)

    @property
    def status(self) -> str:
        """Returns the last result status."""
        last_result_status, _, _ = self._get_status()
        return last_result_status

    @property
    def stdout(self) -> str:
        """Returns the last result stdout."""
        _, last_result_stdout, _ = self._get_status()
        return last_result_stdout

    @property
    def stderr(self) -> str:
        """Returns the last result stderr."""
        _, _, last_result_stderr = self._get_status()
        return last_result_stderr

    def _get_status(self) -> Tuple[str, str, str]:
        """Returns a tuple of (last result status, last result stdout)."""
        db = TinyDB(DB_FILENAME)
        Status = Query()

        values = db.search(Status.id == self._db_id)

        if not values:
            raise ValueError("This check hasn't run yet")

        last_result_status = values[-1]["status"]
        last_result_stdout = values[-1]["stdout"]
        last_result_stderr = values[-1]["stderr"]
        return last_result_status, last_result_stdout, last_result_stderr

    def _set_status(
        self, status: Literal["OK", "NOT OK", "UNKNOWN"], stdout: str, stderr: str
    ) -> None:
        """Sets the last result status and stdout."""
        db = TinyDB(DB_FILENAME)
        Status = Query()

        data = {
            "id": self._db_id,
            "status": status,
            "stdout": stdout,
            "stderr": stderr,
        }

        try:
            self._get_status()
        except ValueError:
            db.insert(data)
            return

        db.update(data, Status.id == self._db_id)

    @property
    def _db_id(self) -> str:
        return f"{self.host}:{self.command}"

    def run(self) -> Tuple[bool, bool, str, str]:
        """Executes the command and checks the results."""

        stdout, stderr = self._execute_ssh_command()

        try:
            last_result_stdout = self.stdout
        except ValueError:
            last_result_stdout = None

        stdout_changed = stdout != last_result_stdout
        status = "OK" if stdout == self.expected else "NOT OK"

        self._set_status(status, stdout, stderr)

        return stdout_changed, stdout == self.expected, stdout, stderr



#!/usr/bin/env python

import atexit
import json
import logging
import os
import pprint
import signal
import sys
import yaml

from typing import Literal

from fikkie.config import BASE_DIR, BROKER_DIR, CONFIG_FILE, LOG_FILE, PID_FILE
from fikkie.main import app
from fikkie.watchdog import WatchDog


CONFIG_TEMPLATE = """---
## SSH config
# Fikkie needs to know which user to use to login into the servers. This will
# default to "fikkie".
#
# Example:
#
# ssh:
#   username: fikkie

## Servers
# This is where you specify the commands fikkie needs to execute over SSH to
# test them.
#
# Example:
#
# servers:
#   primary.foo.com:
#     - description: 'MariaDB'
#       command: 'sudo systemctl status mariadb | grep "Active: active" -c'
#       expected: '1'
#     - description: 'HTTP code foo.com'
#       command: 'curl -s -o /dev/null -w "%{http_code}" foo.com'
#       expected: '200'

## Notifiers
# If you want fikkie to notify state changes/problems, you'll need to specify
# the notifiers here.
#
# Example:
#
# notifiers:
#   - type: telegram
#     token: '1234:abcd'
#     chat_id: 1234
"""


class Commands:
    FORMAT_OPTIONS = ["YAML", "JSON", "JSON-PRETTY"]

    @staticmethod
    def init(*args, **kwargs):
        """Initialize fikkie's workspace."""
        if os.path.isdir(BASE_DIR):
            logging.warning("Fikkie has already been initialized!")
            exit(1)

        directories = [
            BASE_DIR,
            BROKER_DIR,
            os.path.join(BROKER_DIR, "out"),
            os.path.join(BROKER_DIR, "processed"),
        ]
        for directory in directories:
            os.mkdir(directory)

        with open(CONFIG_FILE, "x") as f:
            f.write(CONFIG_TEMPLATE)

    @staticmethod
    def run(loglevel, *args, **kwargs):
        """Start fikkie."""
        app.worker_main(["worker", "-B", "-l", loglevel])

    @staticmethod
    def start(loglevel, *args, **kwargs):
        """
        Start a fikkie daemon.

        It creates the daemon by forking two times, ensuring we can't get a controlling
        TTY. It also changes the directory to BASE_DIR to have a gaurenteed working
        directory. Finally, it redirects stdout and stderr to the log file and writes
        its PID to the PID_FILE.
        """
        if os.path.isfile(PID_FILE):
            logging.warning("PID file exists, is fikkie already running?")
            exit(1)

        if os.fork():
            logging.info("Starting a fikkie daemon.")
            sys.exit()

        os.chdir(BASE_DIR)

        if os.fork():
            sys.exit()

        sys.stderr.flush()
        sys.stdout.flush()
        with open(LOG_FILE, "a+b", 0) as log_file:
            os.dup2(log_file.fileno(), sys.stderr.fileno())
            os.dup2(log_file.fileno(), sys.stdout.fileno())

        def _delete_pid_file():
            os.remove(PID_FILE)

        atexit.register(_delete_pid_file)
        with open(PID_FILE, "w+") as pid_file:
            pid_file.write(f"{os.getpid()}")

        app.worker_main(["worker", "-B", "-l", loglevel])

    @staticmethod
    def stop(*args, **kwargs):
        """Stop the fikkie daemon."""
        try:
            with open(PID_FILE, "r") as pid_file:
                pid = int(pid_file.read().strip())
        except FileNotFoundError:
            logging.error("Could not find PID file. Is fikkie running?")
            exit(1)

        try:
            os.kill(pid, signal.SIGTERM)
            logging.info("Stopped the fikkie daemon.")
        except ProcessLookupError:
            os.remove(PID_FILE)
            logging.warning("The fikkie daemon had already stopped.")

    @staticmethod
    def status(output_format: Literal["YAML", "JSON", "JSON-PRETTY"], *args, **kwargs):
        """Get status from all servers."""

        def _dump(check):
            try:
                return check.dump()
            except ValueError:
                return check.dump(short=True)

        watchdog = WatchDog()
        data = {
            "Daemon running": os.path.isfile(PID_FILE),
            "checks": [_dump(check) for check in watchdog.checks],
        }

        if output_format.upper() == "YAML":
            print(yaml.safe_dump(data))
        elif output_format.upper() == "JSON":
            print(json.dumps(data))
        elif output_format.upper() == "JSON-PRETTY":
            pprint.pprint(data)
        else:
            logging.error(f"Unknown output format: {output_format}")
            exit(1)
